/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

/** @file
 *  @brief Header file containing declarations of functions and data structures for the JPEG-HDR mobile encoder library.
 *
 *  This file contains declarations of the data structures and functions that are used by the JPEG-HDR mobile encoder library.
 */

#ifndef _JH_ENCODER_H
#define _JH_ENCODER_H

#ifdef __cplusplus
extern "C" {
#endif

#include "jh_datatypes.h"
#include "jh_color.h"

/** @defgroup encodint Encoder Interface
 *  @brief Defines the API for the JPEG-HDR mobile encoder library.
 *  @{
 */

#define JH_RI_LUT_SIZE 256                    /* size of ratio_image_lut in jh_encoder_params_t struct */

/** @brief Defines the encoder parameters.
 */
typedef struct jh_encoder_params_ts {
    int base_quality;                           /**< Base image JPEG encoding quality. Valid values 0 to 100, default 95. */
    int subband_quality;                        /**< Subband image JPEG encoding quality. Valid values 0 to 100, default 85. */
    int base_subsampling;                       /**< Base image JPEG subsampling. Valid values 0, 420(same as 0), 422, 444 */
    int subband_subsampling;                    /**< Subband image JPEG subsampling. Valid values 0, 420(same as 0), 422, 444 */
    float samp2nits;                            /**< NIT scaler to write into the header. Use 0 if the value is unknown. */
    float noise_floor;                          /**< Controls the accuracy of encoding, particularly in darker areas. Valid
                                                     values 0 to 1, default 0.0001. The value depends on the camera sensor noise. */
    float chroma_minmax[4];                     /**< Range of Cb and Cr in the residual image { cb_min, cb_max, cr_min, cr_max }.
                                                     The default value of all elements is 0. */
    jh_marker_list_t *markers;                  /**< User-defined container for storing JPEG APP and COM data */
    float cs_conv_mat[JH_CS_MATRIX_SIZE];       /**< Base image color space RGB -> HDR RGB conversion matrix (csb in header) */
    float cs_resi_mat[JH_CS_MATRIX_SIZE];       /**< Color conversion matrix (to HDR RGB) for Cb/Cr residuals (csr in header) */
    float base_cs_trc[JH_CS_TRC_SIZE];          /**< Base image tone reproduction curve */
    float target_cs_mat[JH_CS_MATRIX_SIZE];     /**< Target (HDR) color space matrix (RGB -> XYZ; cst in header) */
    int predictive_coding;                      /**< 1 to use JPEG-compressed base image during encoding; 0 to use original base
                                                     image; default = 0 */
    float swb[3];                               /**< White balance multiplier in camera sensor RGB space. */
    float spr[9];                               /**< Camera sensor to target HDR color space conversion matrix. */
} jh_encoder_params_t;


/** @brief Encodes the HDR data into JPEG-HDR.
 *  @return 0 if the function does not succeed.
 */
int jh_encode(
    const jh_encoder_params_t *encoder_params,  /**< [in]     Encoder parameters. */
    jh_monitor_t *monitor,                      /**< [in]     Progress status of the function. The initial value of the 
                                                              jh_monitor_t @c cancel field must be 0. */
    jh_image_t *hdri,                           /**< [in]     Input HDR image. If specified, the memory can be reused during
                                                              encoding. */
    jh_image_t *sdri,                           /**< [in]     Input tone-mapped SDR image.  */
    jh_stream_t *stream,                        /**< [out]    Output data stream. */
    jh_error_t* error                           /**< [in,out] User-allocated error-reporting structure that contains information
                                                              about the error. */
);

/** @brief Returns the encoder library version in a formatted string of @c major.minor.build.
 *  @return The library version in a null-terminated character string.
 */
const char* jh_encoder_version();

/** @brief Sets the default value of the encoder parameters.
 *  @return 0 if the function does not succeed.
 */
int jh_encoder_set_default(
    jh_encoder_params_t *params,                /**< [out]    User-allocated encoder parameters.  */
    jh_error_t *error                           /**< [in,out] User-allocated error-reporting structure that contains information
                                                              about the error. */
);

/**@}*/  /* end of defgroup encodint */

#ifdef __cplusplus
}
#endif

#endif /*_JH_ENCODER_H*/
