/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "jh_decoder.h"
#include "jh_helpers.h"
#include "jh_malloc.h"
#include "jh_log.h"
#include "jh_marker_list.h"
#include "jh_stream.h"
#include "jh_jpeg_wrap.h"
#include "jh_datatypes.h"
#include "jh_defs.h"
#include "jh_color.h"

typedef struct {
    /*input*/
    jh_stream_t*   stream;         /* input stream with jpeg-hdr */
    
    /*output after jpeg_decompression*/
    unsigned char* base;          /* 8-bit gamma corrected base image YCbCr or BW? gamma*/
    unsigned char* sbi;           /* 8-bit log luminance ratio BW or YCbCr interleaved + 8-bit cb + cr residual*/
    char           header[JH_MAX_HEADER_SIZE];/* ASCII string for JPEG-HDR header */
    unsigned char* icc;           /* ICC data pointer */
    int            icc_size;      /* size of ICC data */
    float          samp2nits;
    
    /*output after jpeg_decode*/
    float*         hdr;           /* reconstructed HDR image in linear RGB */
    int            width;         /* image width */
    int            height;        /* image height */
    void**         markers;        /* user defined container for JPEG APP markers */
    void*          store_segment;
} jh_decoder_struct;

/*Perform JPEG-HDR decoding after JPEG de-compression. It generates HDR image, linear interleaved RGB*/
static jh_result_t jh_decode_hdr(jh_image_t *hdri, jh_image_t *tmi, jh_image_t *sbi, char header[],
                                 jh_error_t* err, jh_monitor_t* monitor);

static char jh_decoder_canceled[] = "HDR decoder canceled";
static char jh_outofmem_msg[] = "Failed to allocate memory in the decoder";


#define JH_JPEGHDR_NUMBER_OF_TAGS 17

typedef struct jh_jpeghdr_header_ts {
    char*   name;       /* Name of the tag in the header */
    int     vector;     /* 1, if tag is associated with an array of NUM FP values (next field), 0 otherwise */
    int     num;        /* When VECTOR is "1", NUM is an array size */
    int     musthave;   /* 0, if parameter is optional. MUSTHAVE becomes 0 on the fly after its tag is parsed */
    float*  data;       /* Pointer to the data holder */
} jh_jpeghdr_header_t;

typedef struct jh_header_params_ts {
    float  lymin;       /* 00 - Log lyminance max */
    float  lymax;       /* 01 - Log lyminance min */
    float  samp2nits;   /* 02 - Sample to nits */
    float  cb_min;      /* 03 - Cb max */
    float  cb_max;      /* 04 - Cb min */
    float  cr_min;      /* 05 - Cr miax */
    float  cr_max;      /* 06 - Cr min */
    float  noise_floor; /* 07 - Noise floor normalized from 0.0 to 1.0 */
    float  y_lut[256];  /* 08 - LUT for decoding ratio */
    float  swb[3];      /* 09 - White balance multiplier in camera sensor RGB space  */
    float  spr[9];      /* 10 - Camera sensor to target HDR color space conversion matrix; see JPEG-HDR spec for details */
    float  cf;          /* 11 - Conversion factor (gain number) for RGB16 color space */
    float  csb[9];      /* 12 - 3x3 matrix in row major form defining linear SDR RGB to linear HDR RGB transformation */
    float  bga[256];    /* 13 - 256-entry lookup table containing the base image gamma-corrected RGB to linear RGB transformation */
    float  ckb;         /* 14 - 4 character checksum; NOTE: This field gets special handling in parser below (it's skipped over). */
    float  csr[9];      /* 15 - 3x3 matrix in row major form defining residual CbCr to RGB transformation */
    float  cst[9];      /* 16 - 3x3 matrix in row major form defining target (HDR) RGB to XYZ transformation */
} jh_header_params_t;

static void jh_init_jpeghdr_header_parser(jh_jpeghdr_header_t* keyTable, jh_header_params_t* params) {
    keyTable[0].name = "ln0";
    keyTable[0].vector = 0;         keyTable[0].num = 0;        keyTable[0].musthave = 0;
    keyTable[0].data = &params->lymin;
    
    keyTable[1].name = "ln1";
    keyTable[1].vector = 0;         keyTable[1].num = 0;        keyTable[1].musthave = 0;
    keyTable[1].data = &params->lymax;
    
    keyTable[2].name = "s2n";
    keyTable[2].vector = 0;         keyTable[2].num = 0;        keyTable[2].musthave = 0;
    keyTable[2].data = &params->samp2nits;
    
    keyTable[3].name = "cb0";
    keyTable[3].vector = 0;         keyTable[3].num = 0;        keyTable[3].musthave = 1;
    keyTable[3].data = &params->cb_min;
    
    keyTable[4].name = "cb1";
    keyTable[4].vector = 0;         keyTable[4].num = 0;        keyTable[4].musthave = 1;
    keyTable[4].data = &params->cb_max;
    
    keyTable[5].name = "cr0";
    keyTable[5].vector = 0;         keyTable[5].num = 0;        keyTable[5].musthave = 1;
    keyTable[5].data = &params->cr_min;
    
    keyTable[6].name = "cr1";
    keyTable[6].vector = 0;         keyTable[6].num = 0;        keyTable[6].musthave = 1;
    keyTable[6].data = &params->cr_max;
    
    keyTable[7].name = "nf";
    keyTable[7].vector = 0;         keyTable[7].num = 0;        keyTable[7].musthave = 1;
    keyTable[7].data = &params->noise_floor;
    
    keyTable[8].name = "riq";
    keyTable[8].vector = 1;         keyTable[8].num = 256;      keyTable[8].musthave = 1;
    keyTable[8].data = &params->y_lut[0];
    
    keyTable[9].name = "swb";
    keyTable[9].vector = 1;         keyTable[9].num = 3;        keyTable[9].musthave = 0;
    keyTable[9].data = &params->swb[0];
    
    keyTable[10].name = "spr";
    keyTable[10].vector = 1;        keyTable[10].num = 9;       keyTable[10].musthave = 0;
    keyTable[10].data = &params->spr[0];
    
    keyTable[11].name = "cf";
    keyTable[11].vector = 0;        keyTable[11].num = 0;       keyTable[11].musthave = 0;
    keyTable[11].data = &params->cf;
    params->cf = -1.0f; /* Gain should be initialized to the only default value different from 0.0 */
    
    keyTable[12].name = "csb";
    keyTable[12].vector = 1;        keyTable[12].num = 9;       keyTable[12].musthave = 0;
    keyTable[12].data = &params->csb[0];
    
    keyTable[13].name = "bga";
    keyTable[13].vector = 1;        keyTable[13].num = 256;       keyTable[13].musthave = 0;
    keyTable[13].data = &params->bga[0];
    
    keyTable[14].name = "ckb";
    keyTable[14].vector = 0;        keyTable[14].num = 0;       keyTable[14].musthave = 0;
    keyTable[14].data = &params->ckb;
    
    keyTable[15].name = "csr";
    keyTable[15].vector = 1;        keyTable[15].num = 9;       keyTable[15].musthave = 0;
    keyTable[15].data = &params->csr[0];
    
    keyTable[16].name = "cst";
    keyTable[16].vector = 1;        keyTable[16].num = 9;       keyTable[16].musthave = 0;
    keyTable[16].data = &params->cst[0];
}

static int jh_lookup_key(jh_jpeghdr_header_t* keyTable, char* key) {
    int i = JH_JPEGHDR_NUMBER_OF_TAGS;
    while (i--) {
        if (strcmp(key, keyTable[i].name) == 0) {
            return i;
        }
    }
    return -1; /* Key not found */
}

static jh_result_t jh_check_must_have_tags(jh_jpeghdr_header_t* keyTable, jh_error_t* err) {
    int i = JH_JPEGHDR_NUMBER_OF_TAGS;
    while (i--) {
        if (keyTable[i].musthave) { /* The tag marked as "must-have" hasn't been parsed */
            if (i != 8) { /* Make "riq" tag (y_lut) an exception */
                jh_set_error(err, JH_M11_FORMAT_ERROR,
                          "the must-have tag %s is absent in Dolby JPEG/HDR header", keyTable[i].name);
                return JH_M11_FORMAT_ERROR;
            }
        }
    }
    return JH_OK;
}

void jh_zero_vector_params(jh_jpeghdr_header_t* keyTable)
{
    int i;
    for(i=0; i<JH_JPEGHDR_NUMBER_OF_TAGS; i++) {
        if(keyTable[i].vector) {
            memset(keyTable[i].data, 0, keyTable[i].num * sizeof(float));
        }
    }
    return;
}

static jh_result_t jh_parse_jpeghdr_header(char* header, jh_jpeghdr_header_t* keyTable, jh_error_t* err) {
    char H[1024];
    char* end = header + strlen(header);
    char e_format_char[128] = {0};
    /* Characters allowed in e-format output are: decimal digits, '.', '-', '+', 'e', 'E' */
    e_format_char['0'] = 1; e_format_char['1'] = 1; e_format_char['2'] = 1; e_format_char['3'] = 1; e_format_char['4'] = 1;
    e_format_char['5'] = 1; e_format_char['6'] = 1; e_format_char['7'] = 1; e_format_char['8'] = 1; e_format_char['9'] = 1;
    e_format_char['.'] = 1; e_format_char['-'] = 1; e_format_char['+'] = 1; e_format_char['e'] = 1; e_format_char['E'] = 1;
    jh_zero_vector_params(keyTable);
    while (header < end) {
        char* ptr = (char*)&H[0];
        int ch, tag;
        if (*header++ != '<') { /* Bad format or garbage in the header */
            jh_set_error(err, JH_M11_FORMAT_ERROR, "Inconsistent Dolby JPEG/HDR header ");
            return JH_M11_FORMAT_ERROR;
        }
        ch = (int)*header++ & 0x7f; /* Parse the tag, which is a character string */
        while ((ch != '=') && (ch != ' ') && (ch != '>')) {
            *ptr++ = (char)ch;
            ch = (int)*header++ & 0x7f;
        }
        if(ch == '>') {
            /* end of a bad tag; continue to next */
            continue;
        }
        *ptr = 0;
        tag = jh_lookup_key(keyTable, H);
        if (tag < 0) { /* Unknown tag; skip to end and continue */
            while (ch != '>') {
                ch = (int)*header++ & 0x7f;
            }
            continue;
        }
        if (keyTable[tag].vector) { /* We have to parse an array of floats */
            int i = keyTable[tag].num; /* Get the number of values in the vector */
            float* fptr = (float*)keyTable[tag].data;
            while (i--) {
                ptr = (char*)&H[0];
                ch = (int)*header++ & 0x7f; /* Parse floating number in either %e or %f format. The break is ' ' or '>' */
                while (e_format_char[ch] && (ch != '>') && (ch != ' ')) {
                    *ptr++ = (char)ch;
                    ch = (int)*header++ & 0x7f;
                }
                *ptr = 0;
                if (sscanf(H, "%f", fptr) == 0) {
                    jh_set_error(err, JH_M11_FORMAT_ERROR,
                              "Unable to parse a vector value of tag %s in Dolby JPEG/HDR header", keyTable[tag].name);
                    return JH_M11_FORMAT_ERROR;
                }
                fptr++;
            }
            if (ch != '>') { /* Vector value of the tag has to close with '>' */
                jh_set_error(err, JH_M11_FORMAT_ERROR,
                          "Vector value of tag %s must contain %d elements", keyTable[tag].name, keyTable[tag].num);
                return JH_M11_FORMAT_ERROR;
            }
            keyTable[tag].musthave = 0; /* Mark that tag was parsed */
        } else { /* Single FP value  */
            if(14 == tag) {
                /* skip over the checksum here */
                header += 5;  /* 4 byte checksum plus the closing '>' */
                continue;
            }
            ptr = (char*)&H[0];
            ch = (int)*header++ & 0x7f; /* Parse floating number in either %e or %f format. The break is '>' */
            while (e_format_char[ch] && ch != '>') {
                *ptr++ = (char)ch;
                ch = (int)*header++ & 0x7f;
            }
            *ptr = 0;
            if (sscanf(H, "%f", (float*)keyTable[tag].data)) {
                keyTable[tag].musthave = 0; /* Mark that tag was parsed */
                continue;
            }
            jh_set_error(err, JH_M11_FORMAT_ERROR,
                      "Unable to parse a value of tag %s in Dolby JPEG/HDR header", keyTable[tag].name);
            return JH_M11_FORMAT_ERROR;
        }
    }
    return jh_check_must_have_tags(keyTable, err);
}

static int jh_valid_params(jh_header_params_t* params) {
    float
    lymin  = params->lymin,
    lymax  = params->lymax,
    samp2nits = params->samp2nits,
    cb_min = params->cb_min,
    cb_max = params->cb_max,
    cr_min = params->cr_min,
    cr_max = params->cr_max,
    noise_floor = params->noise_floor;
    
    if (lymin > 40.0f || lymin < -40.0f || lymin == lymax
        || lymax > 40.0f || lymax < -40.0f
        || samp2nits < 0.f
        || cb_min < -10.f || cb_min > 10.f || cb_min > cb_max
        || cb_max < -10.f || cb_max > 10.f
        || cr_min < -10.f || cr_min > 10.f || cr_min > cr_max
        || cr_max < -10.f || cr_max > 10.f
        || noise_floor < 0.f || noise_floor > 1.f) {
        return 0;
    }
    return 1;
}

/*Perform JPEG-HDR decoding after JPEG de-compression. It generates HDR image, linear interleaved RGB */
static jh_result_t jh_decode_hdr(jh_image_t *hdri, jh_image_t *tmi, jh_image_t *sbi, char header[],
                                 jh_error_t* err, jh_monitor_t* monitor) {
    int header_param_csr_present = 0;
    jh_result_t res = JH_OK;
    int in_header, i, j, channels;
    jh_header_params_t params = {0};
    jh_jpeghdr_header_t keyTable[JH_JPEGHDR_NUMBER_OF_TAGS];
    jh_colorconvert_params_t csp;
    float lymin, lymax, samp2nits, cb_min, cb_max, cr_min, cr_max, noise_floor;
    float
        inverse_csr[JH_CS_MATRIX_SIZE],
        cb_lut[256],  /* subband decoding table */
        cr_lut[256],  /* subband decoding table */
        *lum_coefs = NULL,
        *hdr = NULL;
    unsigned char
    *sbi_data = NULL,
    *tmi_data = NULL;
    memset(&csp, 0, sizeof(jh_colorconvert_params_t));
    jh_init_jpeghdr_header_parser((jh_jpeghdr_header_t*)&keyTable, &params); /* Initialize Parsing Table */
    res = jh_parse_jpeghdr_header(header, (jh_jpeghdr_header_t*)&keyTable, err); /* Parse the header */
    if (res != JH_OK) {
        return res;
    }
    if (!jh_valid_params(&params)) {
        jh_set_error(err, JH_M11_FORMAT_ERROR, "APP11 Marker header contains invalid values");
        return JH_M11_FORMAT_ERROR;
    }
    switch (tmi->pixel_type) {
        default:
        case JH_RGBF:
        case JH_GRAYF:
            jh_set_error(err, JH_UNSUPPORTED, "Only 8-bit tonemapped and subband images are supported");
            return JH_UNSUPPORTED;
            break;
        case JH_RGB8:
            channels = 3;
            break;
        case JH_GRAY8:
            channels = 1;
            break;
    };
    if(jh_channel_count(tmi->pixel_type) != jh_channel_count(sbi->pixel_type)) {
        jh_set_error(err, JH_UNSUPPORTED, "Tonemapped and subband image must have the same number of channels.");
        return JH_UNSUPPORTED;
    }
    jh_init_colorconvert_params(&csp, params.csb, params.bga);
    lymin  = params.lymin,
    lymax  = params.lymax,
    samp2nits = params.samp2nits,
    cb_min = params.cb_min,
    cb_max = params.cb_max,
    cr_min = params.cr_min,
    cr_max = params.cr_max,
    noise_floor = params.noise_floor;
    
    if (params.cf > 1.f) { /* the gain is set */
        hdri->decoding_lut = (float *) jh_malloc(sizeof(float));
        if (NULL == hdri->decoding_lut) {
            res = JH_MEMORY;
            jh_set_error(err, JH_MEMORY, jh_outofmem_msg);
            goto cleanup;
        }
        hdri->decoding_lut[0] = params.cf;
        hdri->lut_size = 1;
    }
    in_header = keyTable[8].musthave ? 0 : 1; /* 1, if <riq ...> tag was parsed in the header */
    if (!in_header) { /* No lut in header */
        for (i = 0; i < 256; i++) {
            params.y_lut[i] = powf(2.f, ((i*(lymax - lymin))/(255.0f) + lymin));
        }
    }
    for (i = 0; i < 256; i++) {
        cb_lut[i] = (i * (cb_max - cb_min)) / 255.f + cb_min;
        cr_lut[i] = (i * (cr_max - cr_min)) / 255.f + cr_min;
    }
    for(i=0; i<JH_CS_MATRIX_SIZE; i++) {
        if(fabs(params.csr[i]) >= JH_MIN_FLOAT) {
            header_param_csr_present = 1;
            break;
        }
    }
    if((header_param_csr_present) && (0 == jh_invert_3x3_matrix(params.csr, inverse_csr))) {
        /* if csr is in header and it is invertible, use it for jh_hdr_lum coefficients.  Otherwise lum_coefs is NULL and 
           default coefficients are used. */
        lum_coefs = inverse_csr;  /* first 3 entries are Y coefficients */
    }
    if(3 == channels) {
        for(i = 0; i < hdri->rows; i++) {
            sbi_data = (unsigned char*) jh_get_row(sbi, i);
            tmi_data = (unsigned char*) jh_get_row(tmi, i);
            hdr = (float*) jh_get_row(hdri, i);
            for (j = 0; j < hdri->columns * channels; j += channels) {
                float ratio = params.y_lut[sbi_data[j]];
                float rgb[3], scale, cb, cr, resR, resG, resB;
                jh_baseRGB_to_hdrRGB(&(tmi_data[j]), rgb, &csp);
                scale = jh_hdr_lum(rgb[0], rgb[1], rgb[2], lum_coefs) + noise_floor;
                cb = cb_lut[sbi_data[j+1]] * scale;
                cr = cr_lut[sbi_data[j+2]] * scale;
                if(header_param_csr_present) {
                    resR = params.csr[1] * cb + params.csr[2] * cr;
                    resG = params.csr[4] * cb + params.csr[5] * cr;
                    resB = params.csr[7] * cb + params.csr[8] * cr;
                }
                else {
                    resR = (                  1.402f  * cr);
                    resG = (-0.34414f * cb - 0.71414f * cr);
                    resB = ( 1.772f   * cb                );
                }
                hdr[j  ] = ratio * (rgb[0] + resR); hdr[j  ] = hdr[j  ] < 0.f ? 0.f : hdr[j  ];
                hdr[j+1] = ratio * (rgb[1] + resG); hdr[j+1] = hdr[j+1] < 0.f ? 0.f : hdr[j+1];
                hdr[j+2] = ratio * (rgb[2] + resB); hdr[j+2] = hdr[j+2] < 0.f ? 0.f : hdr[j+2];
                if (monitor && jh_test_and_reset(&monitor->cancel)) {
                    jh_set_error(err, JH_CANCELED, jh_decoder_canceled);
                    return JH_CANCELED;
                }
            }
        }
    }
    else if(1 == channels) {
        for(i = 0; i < hdri->rows; i++) {
            sbi_data = (unsigned char*) jh_get_row(sbi, i);
            tmi_data = (unsigned char*) jh_get_row(tmi, i);
            hdr = (float*) jh_get_row(hdri, i);
            for (j = 0; j < hdri->columns; j++) {
                float ratio = params.y_lut[sbi_data[j]];
                float rgb[3], lum;
                unsigned char tmi[3];
                tmi[0] = (tmi[1] = (tmi[2] = tmi_data[j]));
                jh_baseRGB_to_hdrRGB(tmi, rgb, &csp);
                lum = jh_hdr_lum(rgb[0], rgb[1], rgb[2], lum_coefs);
                hdr[j] = ratio * lum;
                if (monitor && jh_test_and_reset(&monitor->cancel)) {
                    jh_set_error(err, JH_CANCELED, jh_decoder_canceled);
                    return JH_CANCELED;
                }
            }
        }
    }
cleanup:
    if (JH_OK != res) { /* on error exit */
        jh_delete_image(hdri);
    }
    return res;
}

static int jh_store_dolby_markers(jh_marker_list_t* markers, jh_mem_stream_t* sbi_strm, char header[],
        const int header_size, jh_marker_segment_t* dolby_segs[], int probe, jh_error_t* err);

int jh_is_hdr(jh_stream_t *strm) {
    int ret = 1;
    jh_error_t err = {JH_OK};
    jh_mem_stream_t sbi_strm = {0};
    jh_buf_mgr_t sbi_mgr = {0};
    jh_marker_segment_t* dolby_segs[JH_MAX_NO_SEGS] = {0};
    jh_marker_list_impl_t local_markers = {{0}};
    if (NULL == strm) {
        ret = 0;
        goto cleanup;
    }
    jh_init_stream(&sbi_strm, &sbi_mgr); /* init stream for sbi */
    jh_init_marker_list(&local_markers);
    ret = strm->seek(strm, 0, JH_SEEK_SET);
    if (!ret) {
        goto cleanup;
    }
    ret = jh_jpeg_extract_markers(strm, (jh_marker_list_t*)&local_markers, &err);
    if (!ret) { /* extract all markers except 0 and 1*/
        goto cleanup;
    }
    ret = jh_store_dolby_markers((jh_marker_list_t*)&local_markers, &sbi_strm, NULL,
            JH_MAX_HEADER_SIZE, dolby_segs, 1, &err);
    if (!ret) {  /* store Dolby markers in a stream */
        goto cleanup;
    }
cleanup:
    if(local_markers.cur) {
        local_markers.pub.empty((jh_marker_list_t*)&local_markers);
    }
    return ret;
}

int jh_get_size(jh_stream_t*  stream, int* width, int* height, int* channels, jh_error_t* error) {
    if (jh_is_invalid_error(error)
        || jh_is_null(stream, "stream", error)
        || jh_is_null(width, "width", error)
        || jh_is_null(height, "height", error)
        || jh_is_null(channels, "channels", error)) {
        return 0;
    }
    return jh_jpeg_read_header(stream, width, height, channels, error);
}


/* None of the decoder parameters have strictly defined maxima so we only check
   minimums here. */

#define JH_CHECK_DEC_PARAM_RANGE(param, pmin, err, err_format_str)   \
        if (param < pmin) {                         \
            jh_set_error(err, JH_INVALID_ARGUMENT, err_format_str, param); \
            return 1;                                                   \
        }                                                               \

static int jh_is_equal_check(size_t val, size_t eq_val, jh_error_t *err, char* msg) {
    if (val == eq_val) {
        jh_set_error(err, JH_INVALID_ARGUMENT, msg, eq_val);
        return 1;
    }
    return 0;
}
static int jh_is_invalid_dec_params(const jh_decoder_params_t * p, jh_error_t *err) {
    if (p == NULL) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "NULL decoder parameters");
        return 1;
    }
    JH_CHECK_DEC_PARAM_RANGE(p->num, 1, err, "num value %d must be greater than or equal to 1")
    JH_CHECK_DEC_PARAM_RANGE(p->denom, 1, err, "denom value %d must be greater than or equal to 1")
    if (
        jh_is_equal_check((size_t)p->marker_list, 0, err, "jh_decode: marker_list should not be NULL") 
        || jh_is_equal_check((size_t)p->marker_list->add, 0, err, "jh_decode: marker_list interface should be initialized (add)")
        || jh_is_equal_check((size_t)p->marker_list->empty, 0, err, "jh_decode: marker_list interface should be initialized (empty)")
        || jh_is_equal_check((size_t)p->marker_list->jh_get_prev, 0, err, "jh_decode: marker_list interface should be initialized (jh_get_prev)")
        || jh_is_equal_check((size_t)p->marker_list->jh_get_next, 0, err, "jh_decode: marker_list interface should be initialized (jh_get_next)")
        || jh_is_equal_check((size_t)p->marker_list->get, 0, err, "jh_decode: marker_list interface should be initialized (get)")) {
            return 1;
    }
    return 0;
}

#undef JH_CHECK_DEC_PARAM_RANGE

static const char app11_version_error[] = "Wrong version of JPEG_HDR APP11 header";
static const unsigned char parameter_signature[] = {'D', 'D', 0x00, 0x01}; /** parameter segment signature */
static const unsigned char residual_signature[] = {'D', 'D', 0x00, 0x02}; /** residual segment signature */
static const unsigned int *param_sig = (const unsigned int*)parameter_signature;
static const unsigned int *res_sig = (const unsigned int*)residual_signature;

static int jh_store_dolby_markers(jh_marker_list_t* markers, jh_mem_stream_t* sbi_strm, char header[],
        const int header_size, jh_marker_segment_t* dolby_segs[], int probe, jh_error_t* err) {
    jh_marker_segment_t *seg = NULL;
    jh_node_t *cur_node = NULL;
    const unsigned int *sig = NULL;
    unsigned char 
        *data = NULL,
        *m11[JH_MAX_NO_SEGS] = {NULL};
    int
        size = 0,
        cur = 0,                /* current byte count in the buffer */
        i = 0,
        count = 0,              /* number of DD HDR segment found */
        m11_noseg = 0,          /* number of DD HDR ap11 segments */
        m11_sizes[JH_MAX_NO_SEGS] = {0};          /* sizes of buffers from m11[] */;

    /* Extract and sort dolby segments */
    while(markers->jh_get_prev(markers, &seg));  /* to the beginning */
    if (!markers->get(markers, &seg)) {
        jh_set_error(err, JH_JPEG_ERROR, "No JPEG markers found");
        return 0;
    }
    do {
        if (JH_APPM != seg->marker) {
            continue;
        }
        cur = 4; /* current byte number set after the signature */
        size = seg->length;
        data = seg->data;
        sig = (const unsigned int*)data;
        if (*sig == *param_sig) { /* Dolby's HDR Parameter segment */
            unsigned noseg = 0;
            if ((data[cur] != JH_FORMAT_VERSION_MAJOR) ||
                (data[cur+1] != JH_FORMAT_VERSION_MINOR)) {
                jh_set_error(err, JH_M11_VERSION_ERROR, app11_version_error);
                return 0; /* Current version is 2.1. Used to be 2.0 */
            }
            cur += 2;
            noseg = data[cur++] << 8;
            noseg |= data[cur++];
            size -= cur; /* 'size' is now a size of HDR parameters data */
            if ((noseg == 0) || (size >= JH_MAX_HEADER_SIZE)) {
                jh_set_error(err, JH_M11_FORMAT_ERROR, "Number of segments out of range. %d segments reported in header", noseg);
                return 0;
            }
            m11_noseg = noseg;
            if (noseg > JH_MAX_NO_SEGS) { /*we have to increase memory allocated for HDR m11*/
                jh_set_error(err, JH_NOT_IMPLEMENTED, "This version does not support more than %d segments", JH_MAX_NO_SEGS);
                return 0;
            }
            if (!probe) { /* We've identified JPEG-HDR format */
                if (size <= header_size) {
                    memcpy(header, &data[cur], size);
                    dolby_segs[count++] = seg; /* store Dolby segments for future removal */
                } else {
                    jh_set_error(err, JH_M11_FORMAT_ERROR, "Header size (%d bytes) exceeds the limit for this version", size);
                    return 0;
                }
            } else {
                return noseg > 0;  /* probing thus don't need to go further */
            }
        } else if (*sig == *res_sig) { /* DD HDR residual segment identified */
            unsigned segno = 0;
            if (probe) { /* Skip residual segment because we are only probing header */
                continue;
            }
            segno = data[cur++] << 8;
            segno |= data[cur++];
            if (segno >= JH_MAX_NO_SEGS) {  /* this reallocation handles cases that when header segment is not the first */
                jh_set_error(err, JH_NOT_IMPLEMENTED, "This version does not support more than %d segments", JH_MAX_NO_SEGS);
                return 0;
            }
            size -= cur; /* 'size' is now a size of ap11 residual image data */
            m11_sizes[segno] = size;
            m11[segno] = &data[cur];
            dolby_segs[count++] = seg; /* store Dolby segments for future removal */
        }
    } while (markers->jh_get_next(markers, &seg));
    /* fill the stream with data */
    for (i = m11_noseg - 1 ; i >= 0 ; i-- ) {
        cur_node = jh_create_node(m11_sizes[i], m11[i], NULL, cur_node);
    }
    sbi_strm->current = sbi_strm->head = cur_node;
    return m11_noseg > 0;
}

int jh_decoder_set_default(jh_decoder_params_t *dp, jh_error_t *err)
{
    if (NULL == err) {
        return 0;
    }
    if (NULL == dp) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "jh_decoder_set_default: null pointer passed as parameter");
        return 0;
    }
    dp->num = 1;
    dp->denom = 1;
    dp->marker_list = (jh_marker_list_t *) 0;
    return 1;
}

jh_image_t* jh_decode(jh_stream_t *strm, jh_monitor_t *monitor,
                      const jh_decoder_params_t *dp, jh_error_t* error) {
    jh_result_t res = JH_ERROR;
    char header[JH_MAX_HEADER_SIZE] = {0};
    jh_image_t* image = NULL;
    jh_image_t
            tmi = {0},
            sbi = {0};
    jh_mem_stream_t sbi_strm = {0};
    jh_buf_mgr_t sbi_mgr = {0};
    jh_marker_segment_t* dolby_segs[JH_MAX_NO_SEGS] = {0};
    int i = 0,
            width=0,
            height = 0,
            channels = 0;
    if (jh_is_invalid_error(error)
        || jh_is_null(strm, "stream", error)
        || jh_is_null(monitor, "monitor", error)
        || jh_is_invalid_dec_params(dp, error)) {
        return NULL;
    }
    jh_init_stream(&sbi_strm, &sbi_mgr); /* init stream for sbi */
    strm->seek(strm, 0, JH_SEEK_SET);
    if (!jh_jpeg_extract_markers(strm, dp->marker_list, error)) { /* extract all markers except 0 and 1*/
        res = error->code;
        goto cleanup;
    }
    if (!jh_store_dolby_markers(dp->marker_list, &sbi_strm, header,
            JH_MAX_HEADER_SIZE, dolby_segs, 0, error)) {  /* store Dolby markers in a stream */
        res = error->code;
        goto cleanup;
    }
    sbi_strm.seek((jh_stream_t*)&sbi_strm, 0, JH_SEEK_SET);
    if(!jh_get_size((jh_stream_t*)&sbi_strm, &width, &height, &channels, error)) {  /* need to know channel count; get from header */
        res = error->code;
        goto cleanup;
    }
    sbi_strm.seek((jh_stream_t*)&sbi_strm, 0, JH_SEEK_SET);
    sbi.pixel_type = (3 == channels) ? JH_YUV444I8 : JH_GRAY8;
    if (!jh_jpeg_extract_image((jh_stream_t*)&sbi_strm, dp->num, dp->denom, &sbi, error)) {
        res = error->code;
        goto cleanup;
    }
    tmi.pixel_type = (3 == channels) ? JH_RGB8 : JH_GRAY8;
    strm->seek(strm, 0, JH_SEEK_SET);
    if (!jh_jpeg_extract_image(strm, dp->num, dp->denom, &tmi, error)) {
        res = error->code;
        goto cleanup;
    }
    image = jh_create_image(tmi.columns, tmi.rows, (3 == channels) ? JH_RGBF : JH_GRAYF);
    if (NULL == image) {
        jh_set_error(error, JH_MEMORY, jh_outofmem_msg);
        res = JH_MEMORY;
        goto cleanup;
    }
    res = jh_decode_hdr(image, &tmi, &sbi, header, error, monitor);
    if (JH_OK != res) {
        goto cleanup;
    }
cleanup:
    if (JH_OK != res) {
        jh_delete_image(image);
        image = NULL;
    } else {  /* delete Dolby segments if processing was successful */
        i = 0;
        while(dp->marker_list->jh_delete_segment(dp->marker_list, &dolby_segs[i]) && i < JH_MAX_NO_SEGS) {
            ++i;
        }
    }
    sbi_strm.empty((jh_stream_t*)&sbi_strm);
    jh_jpeg_release_image_data((unsigned char *) tmi.data);
    jh_jpeg_release_image_data((unsigned char *) sbi.data);
    return image;
}

const char* jh_decoder_version() {
   static const char decoder_version[] = JH_LIB_VERSION;
   return decoder_version;
}
