/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "jh_helpers.h"
#include "jh_encoder.h"
#include "jh_jpeg_wrap.h"
#include "jh_log.h"
#include "jh_marker_list.h"
#include "jh_stream.h"
#include "jh_color.h"
#include "jh_defs.h"

#define JH_MAX_MARKER_DATA_SIZE (0xFFFF - 2)
#define JH_INV_LUT_MAXSIZE ((1<<20) - 1)

static int jh_calculate_residual_image(const jh_encoder_params_t *enc_params,
                        jh_monitor_t *monitor, jh_image_t *hdri,
                        jh_image_t *sdri, char *header, jh_error_t *err);

static int jh_create_jpeghdr(const char header[], const jh_image_t *sdri,
                          const jh_encoder_params_t *params, jh_stream_t *fstrm,
                          jh_monitor_t * monitor, jh_error_t *err);


static int jh_add_dolby_data(const char * header, jh_stream_t* strm,
                          size_t sbi_jpg_size,
                          jh_marker_list_t *markers, jh_error_t *err) {
    int
    j = (int)sbi_jpg_size,
    k = (int)strlen((const char*)header);
    const int
    mh_size = (int)strlen(JH_ID_STRING) + 6,    /* master header size =
                                                 0x0+type(1b)+ver(2b)+total(2b) */
    ch_size = mh_size - 2;                      /* child hearder size  =
                                                 0x0 + type(1b) + seg#(2b)*/
    unsigned char sig[64] = {0};            /* dobly segment signature and info */
    int nsegs = 0, ret = 1;
    jh_marker_segment_t *segment = NULL;
    /* Calculate memory requirements for M11.
     * M11 structure header segment - full child segments - partial child segment (if exist) */
    nsegs = (int)((float)j/(JH_MAX_MARKER_DATA_SIZE - ch_size)); /* # child segments */
    nsegs += 1 + (j%(JH_MAX_MARKER_DATA_SIZE - ch_size) > 0); /* add header segment and the last segment if exist */
    if (JH_MAX_NO_SEGS < nsegs) {
        jh_set_error(err, JH_ENCODER_ERROR, "Too many segments required to encode image");
        goto cleanup;
    }
    /* create header segment */
    strncpy((char*)sig, JH_ID_STRING, 64);
    sig[mh_size-5] = (unsigned char)1;
    sig[mh_size-4] = (unsigned char)JH_FORMAT_VERSION_MAJOR;
    sig[mh_size-3] = (unsigned char)JH_FORMAT_VERSION_MINOR;
    sig[mh_size-2] = (unsigned char)((nsegs-1)/256);
    sig[mh_size-1] = (unsigned char)((nsegs-1)%256);
    ret = markers->add((jh_marker_list_t*)markers, JH_APPM, mh_size, sig, k, (unsigned char*)header);
    if (!ret) {
        jh_set_error(err, JH_ENCODER_ERROR, "Failed to add HDR metadata header to markers");
        goto cleanup;
    }
    /* add the rest of the data */
    for (j = 1; j < nsegs; j++){
        strncpy((char*)sig, JH_ID_STRING, 64);
        sig[ch_size-3] = (unsigned char)2;
        /*segment index 0 based*/
        sig[ch_size-2] = (unsigned char)((j-1)/256);
        sig[ch_size-1] = (unsigned char)((j-1)%256);
        k = JH_MAX_MARKER_DATA_SIZE - ch_size;
        k = k > (int)sbi_jpg_size ? (int)sbi_jpg_size : k;
        ret = markers->add((jh_marker_list_t*)markers, JH_APPM, ch_size , sig, k, NULL);
        if (!ret) {
            jh_set_error(err, JH_ENCODER_ERROR, "Failed to add HDR metadata to markers");
            goto cleanup;
        }
        ret = markers->get(markers, &segment);
        if (!ret) {
            jh_set_error(err, JH_ENCODER_ERROR, "Failed to get a segment");
            goto cleanup;
        }
        if( k != strm->read(strm, &segment->data[ch_size], k)) {
            ret = 0;
            jh_set_error(err, JH_ENCODER_ERROR, "Failed to read metadata");
            goto cleanup;
        }
        sbi_jpg_size -= k;
    }
cleanup:
    return ret;
}

static int jh_create_jpeghdr(const char header[], const jh_image_t *sdri,
                          const jh_encoder_params_t *params, jh_stream_t *fstrm,
                          jh_monitor_t * monitor, jh_error_t *err) {
    size_t   sbi_jpg_size = 0;
    int
    ret = 1,
    use_local_markers = NULL == params->markers;
    jh_image_t im = *sdri; /* just to clone all image params, but data will be different*/
    jh_mem_stream_t strm = {0};
    jh_buf_mgr_t mgr = {0};
    jh_marker_list_impl_t local_markers = {{0}};
    jh_marker_list_t *markers = use_local_markers
    ? (jh_marker_list_t*)&local_markers
    : params->markers;
    /* stream initialization */
    jh_init_stream(&strm, &mgr);
    jh_init_marker_list(&local_markers);
    /*compress sbi*/
    /* RI data stored in loglum. The RI has the same extensions but YUV type (for RGB images) */
    im.data = sdri->loglum;
    im.pixel_type = (JH_GRAY8 != sdri->pixel_type) ? JH_YUV444I8 : JH_GRAY8;
    ret = jh_jpeg_compress_and_attach(&im, params->subband_quality,
                                      params->subband_subsampling, NULL,
                                      (jh_stream_t*) &strm, monitor, err);
    if(!ret) {
        goto cleanup;
    }
    sbi_jpg_size = strm.get_size((jh_stream_t *)&strm);
    /*create markers, ready to be embedded*/
    ret = strm.seek((jh_stream_t *)&strm, 0, JH_SEEK_SET);
    if (!ret) {
        jh_set_error(err, JH_ENCODER_ERROR, "Failed to rewind internal stream");
    }
    ret = jh_add_dolby_data(header, (jh_stream_t*)&strm, sbi_jpg_size, markers, err);
    if (!ret) {
        goto cleanup;
    }
    /*compress all*/
    ret = jh_jpeg_compress_and_attach(sdri, params->base_quality,
                                      params->base_subsampling, markers, fstrm, monitor, err);
    if (!ret) {
        goto cleanup;
    }
cleanup:
    local_markers.pub.empty((jh_marker_list_t*)&local_markers);
    strm.empty((jh_stream_t*)&strm);
    return ret;
}


static char ri_canceled[] = "jh_calculate_residual_image: Canceled by user";


/* Check, if array of float values should be neglected because every single
 element is less than JH_MIN_FLOAT. SIZE is positive, otherwise it will crash. */
static int jh_count_in_float_array(float const* array, int size) {
    while (size--) {
        if (*array++ >= JH_MIN_FLOAT) {
            return 1; /* We will have to write this array into JPEG-HDR header. */
        }
    }
    return 0; /* All the values are small enough (or negative) to ignore array. */
}

static char smallHeader[] = "Memory allocated for JPEG/HDR header is not enough";

/* For a monotonically increasing lut of length n, return the index i such
 * such that, for value x, lut[i] <= x <= lut[i+1].  If x is outside
 * [0,n-1], then return values is an index of nearest value i.e 0 or n-1.
 */
static int jh_locate(const float *lut, const int n, const float x)
{
    int  pivot, lower = 0, upper = n-1;
    while (upper-lower > 1) {
        pivot=(upper+lower) >> 1;
        if (x >= lut[pivot]) {
            lower=pivot;
        } else {
            upper=pivot;
        }
    }
    return lower;
}

/* Invert a tabulated monotonic function with values in range [mn, mx].
 * The table is first normalized to [0,1] and then inverted.  The result
 * is scaled by 255 to avoid a multiplication when applied to data.
 */
static int jh_invert_table(const float in[], const size_t in_size, float out[], unsigned int *out_size) {
    int ret = 0;
    size_t i,
        required_size = 0;
    float
        dec_lut_range =  in[in_size-1] - in[0],
        d = dec_lut_range;
    /* find minimum step in the table */
    for (i = 0; i < in_size-1; ++i) {
        float delta = in[i+1] - in[i];
        if (delta <= 0.f) {
            ret = 1; /* invalid arguments; the lut should be monotonic */
            goto cleanup;
        }
        if (delta < d) {
            d = delta;
        }
    }
    /* assuming that lut is monotonic */
    required_size = (size_t)ceil(dec_lut_range/d);
    if (required_size < *out_size) {
        *out_size = (unsigned int) required_size;
    }
    for (i = 0; i < *out_size; ++i) {
        float xi = in[0] + i * (in[in_size -1] - in[0])/(*out_size - 1);
        int ind = jh_locate(in, (int) in_size, xi);
        out[i] = ind + (xi - in[ind])/(in[ind+1] - in[ind]);
    }
cleanup:
    return ret;
}

/* interpolated lookup in lut for x in [0, 1] */
static float jh_table_lookup(const float *lut, const int n, const float x)
{
	float frac;
    int ix = (int) (x * n);
    if(ix >= (n-1)) {
        return(lut[n-1]);
    }
    frac = (x * n) - (float) ix;
    return((1.0f - frac) * lut[ix] + frac * lut[ix+1]);
}

static inline float jh_clip(const float x, const float lower, const float upper)
{
    return( (x < lower) ? lower : ((x > upper) ? upper : x) );
}

/* returns the number of bytes written into the header; 0 if there's not enough room left in the header to hold the array */
static int jh_write_header_array(const char *label,
                              const float *array,
                              const unsigned int num_elements,
                              char *header,
                              jh_error_t* err)
{
    const unsigned int CHARS_PER_NUMBER = 15;  /* 14 characters plus a space per number */
    char F[20], ch;
    size_t current_len = strlen(header);
    size_t label_len = strlen(label);
    int size = JH_MAX_HEADER_SIZE - (int)current_len;
    int i, ret = 0;
    char* ts = header + current_len;
    char* ptr = &F[0];
    if ((CHARS_PER_NUMBER * num_elements + label_len + 2) > size) {  /* +2 for enclosing angle brackets */
        jh_set_error(err, JH_MEMORY, smallHeader);
        goto cleanup;
    }
    snprintf(ptr, CHARS_PER_NUMBER + label_len + 1, "<%s %e", label, array[0]); /* +1 for '<' */
    while ((ch = *ptr++) != 0) {
        *ts++ = ch;
    }
    for (i = 1; i < (num_elements-1); i++) {
        ptr = &F[0];
        snprintf(ptr, CHARS_PER_NUMBER, " %e", array[i]);
        while ((ch = *ptr++) != 0) {
            *ts++ = ch;
        }
    }
    ptr = &F[0];
    snprintf(ptr, CHARS_PER_NUMBER + 1, " %e>", array[num_elements-1]);  /* +1 for '>' */
    while ((*ts++ = *ptr++)) ;  /* extra parens to silence compiler warning */
    ret = size - (JH_MAX_HEADER_SIZE - (int) strlen(header));
cleanup:
    return ret;
}

static int jh_calculate_residual_image(const jh_encoder_params_t *enc_params,
                        jh_monitor_t *monitor, jh_image_t *hdri,
                        jh_image_t *sdri, char *header, jh_error_t* err) {
    int ret = 1;
    unsigned int   i, j, size = hdri->columns*hdri->rows;
    unsigned char *tmi = NULL, tmiRGB[3];
    float    lyratio_lut[JH_RI_LUT_SIZE];
    float    inverse_csr[JH_CS_MATRIX_SIZE], *lum_coefs = NULL;
    float    lum_tm, tmp;
    float    lymax = -100000.f, lymin = 100000.f;
    unsigned char   *lyratio8 = NULL;
    float    *lyratio32 = NULL;
    unsigned char   *cb8 = NULL;
    unsigned char   *cr8 = NULL;
    float    cb_max = -100000.f, cb_min = 100000.f;
    float    cr_max = -100000.f, cr_min = 100000.f;
    float    ln_range, rgb[3];
    int compute_minmax = (enc_params->chroma_minmax[0] == enc_params->chroma_minmax[1]);
    jh_colorconvert_params_t csp;
    memset(&csp, 0, sizeof(jh_colorconvert_params_t));
    if(jh_init_colorconvert_params(&csp, enc_params->cs_conv_mat, enc_params->base_cs_trc)) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "HDR color space matrix not invertible");
        return JH_INVALID_ARGUMENT;
    }
    if(0 == jh_invert_3x3_matrix(enc_params->cs_resi_mat, inverse_csr)) {
        /* if csr is invertible, use it for jh_hdr_lum coefficients and residual conversion. */
        lum_coefs = inverse_csr;  /* first 3 entries are Y coefficients */
    }
    if (NULL == sdri->loglum) { /* luminance was not calculated */
        /* Allocating size+2 to make room for lymin and lymax which are expected to be at the end of sdri->loglum. */
        sdri->loglum = jh_malloc(sizeof(float) * (size + 2)); /* don't delete this memory here */
        if( NULL == sdri->loglum ) {
            ret = 0;
            jh_set_error(err, JH_MEMORY, "memory allocation failed in jh_calculate_residual_image_float");
            goto cleanup;
        }
        lymin = 65535.f;
        lymax = -1.f;
        lyratio32 = (float*)sdri->loglum;
        switch (hdri->pixel_type) {
            case JH_RGBF: {
                for ( i = 0; i < hdri->rows; i++ ) {
                    float *pv = (float*) jh_get_row(hdri, i);   /* jhe->hdr; */
                    tmi = (unsigned char*) jh_get_row(sdri, i); /* jhe->base; */
                    /* calculate luminance RI and min/max values */
                    for ( j = 0; j < hdri->columns; j++, pv += 3, tmi += 3, lyratio32++ ) {
                        pv[0] = (pv[0] >= 0.0f) ? pv[0] : 0.0f;
                        pv[1] = (pv[1] >= 0.0f) ? pv[1] : 0.0f;
                        pv[2] = (pv[2] >= 0.0f) ? pv[2] : 0.0f;
                        tmp = jh_hdr_lum(pv[0], pv[1], pv[2], lum_coefs);
                        jh_baseRGB_to_hdrRGB(tmi, rgb, &csp);
                        lum_tm = jh_hdr_lum(rgb[0], rgb[1], rgb[2], lum_coefs);
                        *lyratio32 = tmp/lum_tm;
                        if( *lyratio32 < lymin ) { lymin = *lyratio32; }
                        if( *lyratio32 > lymax ) { lymax = *lyratio32; }
                        if (jh_test_and_reset(&monitor->cancel)) {
                            ret = 0;
                            jh_set_error(err, JH_CANCELED, ri_canceled);
                            goto cleanup;
                        }
                    }
                }
                break;
            }
            case JH_GRAYF: {
                for ( i = 0; i < hdri->rows; i++ ) {
                    float *pv = (float*) jh_get_row(hdri, i);   /* jhe->hdr; */
                    tmi = (unsigned char*) jh_get_row(sdri, i); /* jhe->base; */
                    /* calculate luminance RI and min/max values */
                    for ( j = 0; j < hdri->columns; j++, pv++, tmi++, lyratio32++ ) {
                        *pv = (*pv >= 0.0f) ? *pv : 0.0f;
                        tmp = jh_hdr_lum(*pv, *pv, *pv, lum_coefs);
                        tmiRGB[0] = (tmiRGB[1] = (tmiRGB[2] = *tmi));
                        jh_baseRGB_to_hdrRGB(tmiRGB, rgb, &csp);
                        lum_tm = jh_hdr_lum(rgb[0], rgb[1], rgb[2], lum_coefs);
                        *lyratio32 = tmp/lum_tm;
                        if( *lyratio32 < lymin ) { lymin = *lyratio32; }
                        if( *lyratio32 > lymax ) { lymax = *lyratio32; }
                        if (jh_test_and_reset(&monitor->cancel)) {
                            ret = 0;
                            jh_set_error(err, JH_CANCELED, ri_canceled);
                            goto cleanup;
                        }
                    }
                }
                break;
            }
            default:
                ret = 0;
                jh_set_error(err, JH_INVALID_ARGUMENT, "unknown or unsupported image format");
                goto cleanup;
        }
        *lyratio32++ = lymin; /* store the min & max at the end of loglum... */
        *lyratio32 = lymax;
        lyratio32 = (float*)sdri->loglum;
        lymin = jh_log2(lymin);  /* ... but log2() values are needed below */
        lymax = jh_log2(lymax);
    } else {
        float dr  = 0;
        lyratio32 = (float*)sdri->loglum;
        lymin = jh_log2(lyratio32[sdri->columns * sdri->rows]);
        lymax = jh_log2(lyratio32[sdri->columns * sdri->rows + 1]);
        dr = lymax - lymin; /* dynamic range in stops */
        if (dr > 24.f || dr <= 0.1f) {
            ret = 0;
            jh_set_error(err, JH_INVALID_ARGUMENT,
                      "Dynamic range of precalculated RI lum equals %f and is out of reasonable range.", dr);
            goto cleanup;
        }
    }
    /* calculate logRI to RI float lut */
    for ( i = 0; i < JH_RI_LUT_SIZE; i++ ) {
        lyratio_lut[i] = powf(2.f, ((i*(lymax - lymin))/(JH_RI_LUT_SIZE - 1.0f) + lymin));
    }

    lyratio8 = (unsigned char*)sdri->loglum; /*reuse loglum the memory */;
    cb_min = (cr_min =  65535.0f);
    cb_max = (cr_max = -65535.0f);
    ln_range = 255.f/(lymax - lymin);
    if(!compute_minmax) {
        cb_min = enc_params->chroma_minmax[0];
        cb_max = enc_params->chroma_minmax[1];
        cr_min = enc_params->chroma_minmax[2];
        cr_max = enc_params->chroma_minmax[3];
    }
    /* calculate and encode logRI  also calc residual */
    switch (hdri->pixel_type) {
        case JH_RGBF: {
            float scale_cb, scale_cr, rgb[3], scale;
            float *cbcr32 = (float *) hdri->data;  /* no jh_get_row() here; cbcr32 is assigned more slowly than res32 advances */
            for(j = 0; j < hdri->rows; j++) {
                float *res32 = (float*) jh_get_row(hdri, j);
                tmi = (unsigned char*) jh_get_row(sdri, j); /*jhe->base; */
                for (i = 0; i < hdri->columns; i++, lyratio8 += 3, lyratio32++, tmi += 3, res32 += 3, cbcr32 += 2) {
                    unsigned char tmp_i = 0;
                    float resR, resG, resB;
                    tmp_i = jh_byteclip((jh_log2(*lyratio32) - lymin)*ln_range + 0.5f);
                    *lyratio8 = tmp_i;
                    jh_baseRGB_to_hdrRGB(tmi, rgb, &csp);  /* residual should be computed from color converted TMI,
                                                              since that's how TMI is treated in the decoder */
                    scale = 1.0f/(jh_hdr_lum(rgb[0], rgb[1], rgb[2], lum_coefs) + enc_params->noise_floor);
                    tmp = 1.f/lyratio_lut[tmp_i];
                    resR = (res32[0]*tmp - rgb[0]);
                    resG = (res32[1]*tmp - rgb[1]);
                    resB = (res32[2]*tmp - rgb[2]);
                    if(lum_coefs != NULL) {
                        cbcr32[0] = scale * (inverse_csr[3] * resR + inverse_csr[4] * resG + inverse_csr[5] * resB);
                        cbcr32[1] = scale * (inverse_csr[6] * resR + inverse_csr[7] * resG + inverse_csr[8] * resB);
                    }
                    if(compute_minmax) {
                        if (cbcr32[0] < cb_min) { cb_min = cbcr32[0]; }
                        if (cbcr32[0] > cb_max) { cb_max = cbcr32[0]; }
                        if (cbcr32[1] < cr_min) { cr_min = cbcr32[1]; }
                        if (cbcr32[1] > cr_max) { cr_max = cbcr32[1]; }
                    }
                    if (jh_test_and_reset(&monitor->cancel)) {
                        ret = 0;
                        jh_set_error(err, JH_CANCELED, ri_canceled);
                        goto cleanup;
                    }
                }
            }
            /* encode RI (subband image) */
            lyratio8 = (unsigned char*)sdri->loglum; /* jhe->sbi */;
            cb8 = lyratio8 + 1;
            cr8 = lyratio8 + 2;
            cbcr32 = (float*)hdri->data;
            scale_cb = (cb_min != cb_max) ? 255.f/(cb_max-cb_min) : 0.0f;
            scale_cr = (cr_min != cr_max) ? 255.f/(cr_max-cr_min) : 0.0f;
            for ( i = 0; i < size; i++, cb8 += 3, cr8 += 3, cbcr32 += 2 ) {
                *cb8 = jh_byteclip((cbcr32[0]-cb_min)*scale_cb + 0.5f);
                *cr8 = jh_byteclip((cbcr32[1]-cr_min)*scale_cr + 0.5f);
            }
            break;
        }
        case JH_GRAYF: {
            for (i = 0; i < size; i++, lyratio8++, lyratio32++) {
                *lyratio8 = jh_byteclip((jh_log2(*lyratio32) - lymin)*ln_range + 0.5f);
                if (jh_test_and_reset(&monitor->cancel)) {
                    ret = 0;
                    jh_set_error(err, JH_CANCELED, ri_canceled);
                    goto cleanup;
                }
            }
            if(compute_minmax) {
                cb_min = (cb_max = (cr_min = (cr_max = 0.0f)));
            }
            break;
        }
        default:
            ret = 0;
            jh_set_error(err, JH_INVALID_ARGUMENT, "unknown or unsupported image format");
            goto cleanup;
    }
    /* Populate JPEG/HDR header. Tag order does not matter. */
    snprintf(header, JH_MAX_HEADER_SIZE, "<ln0=%e><ln1=%e><s2n=%e><cb0=%e><cb1=%e><cr0=%e><cr1=%e><nf=%e><ckb=****>",
             lymin, lymax, enc_params->samp2nits,
             cb_min, cb_max, cr_min, cr_max,
             enc_params->noise_floor);  /* "****" is placeholder for checksum (ckb) */
    /* always write riq */
    if(0 == jh_write_header_array("riq", lyratio_lut, JH_RI_LUT_SIZE, header, err)) {
        ret = 0;
        goto cleanup;
    }
    if (jh_count_in_float_array(&enc_params->swb[0], 3)) {
        if(0 == jh_write_header_array("swb", enc_params->swb, 3, header, err)) {
            ret = 0;
            goto cleanup;
        }
    }
    if (jh_count_in_float_array(&enc_params->spr[0], 9)) {
        if(0 == jh_write_header_array("spr", enc_params->spr, 9, header, err)) {
            ret = 0;
            goto cleanup;
        }
    }
    if (jh_count_in_float_array(csp.conv_matrix, JH_CS_MATRIX_SIZE)) {
        if(0 == jh_write_header_array("csb", csp.conv_matrix, JH_CS_MATRIX_SIZE, header, err)) {
            ret = 0;
            goto cleanup;
        }
    }
    if (jh_count_in_float_array(csp.base_trc, JH_CS_TRC_SIZE)) {
        if(0 == jh_write_header_array("bga", csp.base_trc, JH_CS_TRC_SIZE, header, err)) {
            ret = 0;
            goto cleanup;
        }
    }
    if (jh_count_in_float_array(&enc_params->cs_resi_mat[0], JH_CS_MATRIX_SIZE)) {
        if (0 == jh_write_header_array("csr", enc_params->cs_resi_mat, JH_CS_MATRIX_SIZE, header, err)) {
            ret = 0;
            goto cleanup;
        }
    }
    if (jh_count_in_float_array(&enc_params->target_cs_mat[0], JH_CS_MATRIX_SIZE)) {
        if(0 == jh_write_header_array("cst", enc_params->target_cs_mat, JH_CS_MATRIX_SIZE, header, err)) {
            ret = 0;
            goto cleanup;
        }
    }
cleanup:
    return ret;
}

static int jh_encode_float(const jh_encoder_params_t *encoder_params,
        jh_monitor_t *monitor, jh_image_t *hdri, jh_image_t *sdri,
        jh_stream_t *stream, jh_error_t* err);


#define JH_CHECK_ENC_PARAM_RANGE(param, pmin, pmax, err, err_format_str)   \
    if ((param < pmin) || (param > pmax)) {                                \
        jh_set_error(err, JH_INVALID_ARGUMENT, err_format_str, param);     \
        return 1;                                                          \
    }                                                                      \


#define JH_CHECK_ENC_PARAM_MIN(param, pmin, err, err_format_str)       \
    if (param < pmin) {                                                \
        jh_set_error(err, JH_INVALID_ARGUMENT, err_format_str, param); \
        return 1;                                                      \
    }                                                                  \


static int jh_is_invalid_encoder_params(const jh_encoder_params_t *p, jh_error_t *err) {
    if (p == NULL) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "NULL encoder parameters");
        return 1;
    }
    JH_CHECK_ENC_PARAM_RANGE(p->base_quality, 0, 100, err,
        "base_quality parameter %d is out of range [0, 100]")
    JH_CHECK_ENC_PARAM_RANGE(p->subband_quality, 0, 100, err,
        "subband_quality parameter %d is out of range [0, 100]")
    JH_CHECK_ENC_PARAM_MIN(p->samp2nits, 0.0f, err,
        "samp2nits parameter %f must be greater than or equal to 0.0")
    JH_CHECK_ENC_PARAM_RANGE(p->noise_floor, 0.0f, 1.0f, err,
        "noise_floor parameter %f is out of range [0., 1.]")
    JH_CHECK_ENC_PARAM_RANGE(p->predictive_coding, 0, 1, err,
        "predictive_coding must be 0 or 1")
    return 0;
}

#undef JH_CHECK_ENC_PARAM_RANGE
#undef JH_CHECK_ENC_PARAM_MIN
#undef JH_CHECK_ENC_SUBSAMPLING_PARAM


static char encBadPixelType[] = "jh_encode: Unsupported input pixel type: %d";

int jh_encode(const jh_encoder_params_t *encoder_params, jh_monitor_t *monitor,
        jh_image_t *hdri, jh_image_t *sdri, jh_stream_t *stream,
        jh_error_t* err) {
    int ret = 0;
    if (jh_is_invalid_error(err)
            || jh_is_invalid_encoder_params(encoder_params, err)
            || jh_is_invalid_image(hdri, err) || jh_is_invalid_image(sdri, err)
            || jh_is_null((const void *) stream, "stream", err)
            || jh_is_null((const void *) stream->seek, "stream->seek", err)
            || jh_is_null((const void *) stream->write, "stream->write", err)
            || jh_is_null((const void *) stream->pri, "stream->pri", err)
            || jh_is_null((const void *) stream->pri->on_init, "stream->pri->on_init", err)
            || jh_is_null((const void *) stream->pri->on_empty, "stream->pri->on_empty", err)
            || jh_is_null((const void *) stream->pri->on_term, "stream->pri->on_term", err)
            || jh_is_null((const void *) monitor, "monitor", err)) {
        return 0;
    }
    err->code = JH_OK;
    switch (hdri->pixel_type) {
    case JH_RGBF:
    case JH_GRAYF:
        ret = jh_encode_float(encoder_params, monitor, hdri, sdri, stream, err);
        break;
    default:
        jh_set_error(err, JH_INVALID_ARGUMENT, encBadPixelType, hdri->pixel_type);
        return 0;
    }
    return ret;
}

const char* jh_encoder_version() {
    static const char encoder_version[] = JH_LIB_VERSION;
    return encoder_version;
}

int jh_encoder_set_default(jh_encoder_params_t *ep, jh_error_t *err)
{
    const float default_spr[9] = { 0.f, 0.f, 0.f,   0.f, 0.f, 0.f,   0.f, 0.f, 0.f };
    jh_encoder_params_t z = {0};
    if (NULL == err) {
        return 0;
    }
    if (NULL == ep) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "jh_encoder_set_default: null pointer passed as parameter");
        return 0;
    }
    *ep = z; /* zeroing the input */
    ep->base_quality = 95;
    ep->subband_quality = 85;
    ep->samp2nits = 0.f;
    ep->noise_floor = 0.0001f;
    ep->chroma_minmax[0] = (ep->chroma_minmax[1] = (ep->chroma_minmax[2] = (ep->chroma_minmax[3] = 0.0f)));
    ep->predictive_coding = 1;
    ep->base_subsampling = 420;
    ep->subband_subsampling = 444;
    ep->swb[2] = (ep->swb[1] = (ep->swb[0] = .0f));
    memcpy(ep->spr, default_spr, 9 * sizeof(float));
    memcpy(ep->cs_conv_mat, default_csb, JH_CS_MATRIX_SIZE * sizeof(float));
    memcpy(ep->cs_resi_mat, default_csr, JH_CS_MATRIX_SIZE * sizeof(float));
    memset(ep->target_cs_mat, 0, JH_CS_MATRIX_SIZE * sizeof(float));
    jh_get_default_bga(ep->base_cs_trc);
    return 1;
}

/* Round-trip JPEG compression and decompression applied to sdri. */
static int jh_compress_and_decompress_image(jh_image_t *sdri, const jh_encoder_params_t *params, jh_stream_t *fstrm,
                                            jh_monitor_t * monitor, jh_error_t *err)
{
    int ret = 1;
    jh_mem_stream_t strm = {0};
    jh_buf_mgr_t mgr = {0};
    jh_init_stream(&strm, &mgr);
    /*compress image */
    ret = jh_jpeg_compress_and_attach(sdri, params->base_quality, params->base_subsampling,
            NULL, (jh_stream_t*) &strm, monitor, err);
    if (!ret) {
        goto cleanup;
    }
    /* rewind stream */
    ret = strm.seek((jh_stream_t *)&strm, 0, JH_SEEK_SET);
    if (!ret) {
        jh_set_error(err, JH_ENCODER_ERROR, "Failed to rewind internal stream");
    }
    /* decompress image back into sdri */
    ret = jh_jpeg_extract_image((jh_stream_t *)&strm, 1, 1, sdri, err);
cleanup:
    strm.empty((jh_stream_t*)&strm);
    return ret;
}

static int jh_encode_float(const jh_encoder_params_t *encoder_params,
        jh_monitor_t *monitor, jh_image_t *hdri, jh_image_t *sdri,
        jh_stream_t *stream, jh_error_t* err) {
    int ret = 0;
    char header[JH_MAX_HEADER_SIZE] = { 0x00 };
    if(encoder_params->predictive_coding) {
        /* Predictive coding is wasteful in that the SDR image is compressed twice, once here and again in jh_create_jpeghdr(); one 
         * could compress and save the SDR JPEG markers here and reuse them in a modified version of jh_create_jpegjhdr(), but for 
         * now we are living with the inefficiency. */
        jh_image_t *comp_sdri = jh_copy_image(sdri);
        /* compress and decompress the copy */
        ret = jh_compress_and_decompress_image(comp_sdri, encoder_params, stream, monitor, err);
        if(!ret) {
            goto cleanup;
        }
        /* create the residual image using the round-tripped compressed sdri */
        ret = jh_calculate_residual_image(encoder_params, monitor, hdri, comp_sdri, header, err);
        if(!ret) {
            jh_delete_image(comp_sdri);
            goto cleanup;
        }
        /* the loglum field stores the residual image; copy it from the compressed SDR image to the original */
        if(!sdri->loglum) {
            sdri->loglum = jh_malloc(((sdri->columns * sdri->rows) + 2) * sizeof(float));
            if(!sdri->loglum) {
                jh_delete_image(comp_sdri);
                goto cleanup;
            }
        }
        memcpy(sdri->loglum, comp_sdri->loglum, ((sdri->columns * sdri->rows) + 2) * sizeof(float));
        /* sdri now contains the original SDR raster data and a residual computed from a JPEG-compressed version of it */
        jh_delete_image(comp_sdri);
    }
    else {
        ret = jh_calculate_residual_image(encoder_params, monitor, hdri, sdri, header, err);
    }
    if (!ret) {
        goto cleanup;
    }
    ret = jh_create_jpeghdr(header, sdri, encoder_params, stream, monitor, err);
cleanup:
    return ret;
}
