/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "jpeghdr_operations.h"
#include "jh_helpers.h"

#undef RUN_MEMORY_TEST  /* Define this to run an infinite loop of encoding and decoding operations. */

/* Does the file contain Dolby APP11 markers? */
static bool IsDolbyMarker(FILE *in)
{
    const int APP11 = 0xeb;
    int code = fgetc(in);
    if(APP11 == code) {
        /* check for Dolby's "DD" signature */
        int length = fgetc(in) * 256;
        length += fgetc(in);
        if(length >= 4) {
            if((fgetc(in)=='D') && (fgetc(in)=='D')) {
                return true;
            }
        }
    }
    return false;
}

/* Scan through file, looking for any APP11 "DD" markers.  Returns true immediately after one is found, and false otherwise. */
static bool IsDolbyHDR(const char *jfile)
{
    bool isDolbyHDR = false;
    FILE *in = fopen(jfile, "rb");
    if(in) {
        int c;
        while (!feof(in)) {
            c = fgetc(in);
            if (c == 0xFF) {
                if(IsDolbyMarker(in)) {
                    isDolbyHDR = true;
                    break;
                }
            }
        }
        fclose(in);
    }
    return isDolbyHDR;
}

/* Returns true if file begins with FFD8 and false otherwise. */
static bool IsJPEGFile(const char *infile)
{
    FILE *in = fopen(infile, "rb");
    if(in) {
        if(fgetc(in) == 0xFF) {
            if(fgetc(in) == 0xD8) {
                fclose(in);
                return(true);
            }
        }
        fclose(in);
    }
    return(false);
}

int main(int argc,char **argv)
{
    int ret           = EXIT_FAILURE;
    int quality       = -1;
    int hdrquality    = -1;
    char *tmiFile     = NULL;  /* If not overridden with a command-line argument, the internal exposure-gamma TMO will be used. */
    char *dolbyParametersFile = NULL;
    char *prog = argv[0];
    jh_error_t err;
    
    jh_set_error(&err, JH_OK, "No error");
    
    /* Some parameters may be specified on the command line.  Any provided there will override values in the parameters file. */
    while(argc > 3 && argv[1][0] == '-') {
        if (!strcmp(argv[1],"-q")) {
            if (argv[2] == NULL) {
                jh_set_error(&err, JH_INVALID_ARGUMENT, "-q expects a numeric argument.");
                break;
            }
            quality = atoi(argv[2]);
            argv += 2;
            argc -= 2;
        }
        else if (!strcmp(argv[1],"-Q")) {
            if (argv[2] == NULL) {
                jh_set_error(&err, JH_INVALID_ARGUMENT, "-Q expects a numeric argument.");
                break;
            }
            hdrquality = atoi(argv[2]);
            argv += 2;
            argc -= 2;
        }
        else if (!strcmp(argv[1],"-P")) {
            if (argv[2] == NULL) {
                jh_set_error(&err, JH_INVALID_ARGUMENT, "-P expects a filename argument.");
                break;
            }
            dolbyParametersFile = argv[2];
            argv += 2;
            argc -= 2;
        }
        else if (!strcmp(argv[1],"-t")) {
            if (argv[2] == NULL) {
                jh_set_error(&err, JH_INVALID_ARGUMENT, "-t expects a filename argument.");
                break;
            }
            tmiFile = argv[2];
            argv += 2;
            argc -= 2;
        }
        else {
            jh_set_error(&err, JH_INVALID_ARGUMENT, "Unsupported command line switch %s.",argv[1]);
            break;
        }
    }
    
#ifdef RUN_MEMORY_TEST
     /* Memory test code */
    if (argc != 4) {
        fprintf(stderr,
                "\n"
                "Usage: %s [options] source encoded decoded\n"
                "\n"
                "Source image is encoded as a JPEG-HDR file, which is then decoded.\n"
                "This round trip is repeated in an infinite loop.\n"
                "\n"
                "Options include:\n"
                "-q quality : encodes the input and writes JPEG with the given quality\n"
                "             NOTE: if present, overrides base_quality value in Dolby parameters file.\n"
                "-Q quality : defines the quality for the extension layer\n"
                "             NOTE: if present, overrides subband_quality value in Dolby parameters file.\n"
                "-P file    : (Optional) Use JSON-formatted parameters file.\n"
                "-t image   : (Optional) Use tone-mapped image (in 8-bit PPM format) to override internal tone mapping.\n"
                "\n",
                prog);
        return ret;
    }
    
    if(err.code == JH_OK) {
        JPEGOpParams_t hdrParams;
        hdrParams.base_quality = quality;
        hdrParams.subband_quality = hdrquality;
        hdrParams.parametersFile = dolbyParametersFile;  /* may be NULL */
        unsigned long int iters = 0;
        
        while(1) {
            ++iters;
            Encode(argv[1], argv[2], tmiFile, hdrParams, &err);
            if(err.code != JH_OK) {
                break;
            }
            Decode(argv[2], argv[3], hdrParams, &err);
            if(err.code != JH_OK) {
                break;
            }
        }
    }
#else
    /* Normal operation */
    if (argc != 3) {
        fprintf(stderr,
                "\n"
                "Usage: %s [options] source target\n"
                "\n"
                "If source is a JPEG-HDR file, it will be decoded to target.\n"
                "Otherwise source (in PFM or HDR format) will be encoded as a JPEG-HDR file.\n"
                "\n"
                "Options include:\n"
                "-q quality : encodes the input and writes JPEG with the given quality\n"
                "             NOTE: if present, overrides base_quality value in the parameters file.\n"
                "-Q quality : defines the quality for the extension layer\n"
                "             NOTE: if present, overrides subband_quality value in the parameters file.\n"
                "-P file    : (Optional) Use JSON-formatted parameters file.\n"
                "-t image   : (Optional) Use tone-mapped image (in 8-bit PPM format) to override internal tone mapping.\n"
                "\n",
                prog);
        return ret;
    }
    
    if(err.code == JH_OK) {
        if(IsJPEGFile(argv[1])) {
            if(IsDolbyHDR(argv[1])) {
                /* invoke Dolby decoder */
                JPEGOpParams_t hdrParams;
                hdrParams.base_quality = quality;
                hdrParams.subband_quality = hdrquality;
                hdrParams.parametersFile = dolbyParametersFile;  /* may be NULL */
                Decode(argv[1], argv[2], hdrParams, &err);
            }
        }
        else {
            /* invoke Dolby encoder */
            JPEGOpParams_t hdrParams;
            hdrParams.base_quality = quality;
            hdrParams.subband_quality = hdrquality;
            hdrParams.parametersFile = dolbyParametersFile;  /* may be NULL */
            Encode(argv[1], argv[2], tmiFile, hdrParams, &err);
        }
    }
#endif

    if(err.code != JH_OK) {
        fprintf(stderr, "Error (code %d): %s\n", err.code, err.description);
    }
    else {
        ret = EXIT_SUCCESS;
    }
    return ret;
}
