/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <cstdlib>

#define STRL 132

void swap(float *v)
{
    unsigned char *cptr, tmp;
    cptr = (unsigned char *)v;
    tmp = cptr[0];
    cptr[0]=cptr[3];
    cptr[3]=tmp;
    tmp=cptr[1];
    cptr[1]=cptr[2];
    cptr[2]=tmp;
}

int read_pfm(const char* infile, float** hdrdata, unsigned int* width, unsigned int* height, unsigned int *channels)
{
    FILE *fp = NULL;
    int
        failed = 0,
        need_swap = 0,
        i = 0,
        pixels = 0,
        platform_end = 0,
        xres = 0,
        yres = 0;
    long int
        offset = 0,
        file_sz = 0;
    char str[STRL+1];
    float
        *p = NULL,
        scale_end = 0.f;
    *width = 0;
    *height = 0;
    *channels = 0;
    if (NULL == infile) {
        failed = 1;
        goto cleanup;
    }
    fp = fopen(infile, "rb");
    if (NULL == fp) {
        failed = 1;
        goto cleanup;
    }
    if (fseek(fp, 0L, SEEK_END)) {
        failed = 1;
        goto cleanup;
    }
    file_sz = ftell(fp);
    if (file_sz < 0 || fseek(fp, 0L, SEEK_SET)) {
        failed = 1;
        goto cleanup;
    }
    /*ID line*/
    if (!fscanf(fp, "%s:", str)) {
        failed = 1;
        goto cleanup;
    }
    if (('P' != str[0]) || (('F' != str[1]) && ('f' != str[1]))) {
        failed = 1;
        goto cleanup;
    }
    *channels = ('F' == str[1]) ? 3U : 1U;
    /*Dimensions line*/
    if (!fscanf(fp, "%d %d", &xres, &yres)) {
        failed = 1;
        goto cleanup;
    }
    pixels = xres * yres;
    /*Scale factor/endianess*/
    if (!fscanf(fp, "%f", &scale_end)) {
        failed = 1;
        goto cleanup;
    }
    /*platforn endianess*/
    platform_end = 1;
    if (1 == *(char*)&platform_end) {
        platform_end = -1; /*little endian*/
    }
    if (scale_end * platform_end > 0) {
        need_swap = 0;
    } else {
        need_swap = 1;
    }
    /*alloc*/
    *hdrdata = (float *) malloc(pixels * (*channels) * sizeof(float));
    if (NULL == *hdrdata) {
        failed = 1;
        goto cleanup;
    }
    /*PFM raster*/
    p = *hdrdata;
    /* account for possibly different sizes of whitespaces as well as their numbers after PFM header */
    offset = (file_sz - ftell(fp)) - 4*pixels*(*channels);
    if (offset < 0 || fseek(fp, offset, SEEK_CUR)) {
        failed = 1;
        goto cleanup;
    }
    if (pixels * (*channels)!=fread(p, 4, pixels * (*channels), fp)) {
        failed = 1;
        goto cleanup;
    }
    if (need_swap) {
        for(i = 0; i < pixels * (*channels); i++ ) {
            swap(&p[i]);
        }
    }
    *height = yres;
    *width = xres;
cleanup:
    if (NULL != fp) {
        fclose(fp);
    }
    if (failed && (NULL != *hdrdata)) {
        free(*hdrdata);
        *hdrdata = NULL;
    }
    return failed;
}

int write_pfm(const char* outfile,
              float* img,                  /* GRAY or interleaved RGB or RGB float image data  */
              const unsigned int stride,   /* 1 for gray, 3 for RGB                            */
              const unsigned int uwidth,   /* image dimensions                                 */
              const unsigned int uheight)
{
    int ret = 0, platform_end = 1, byte_stride;
    unsigned int y = 0;
    FILE *fp = NULL;
    if (NULL == outfile || NULL == img ) {
        ret = 1;
        goto cleanup;
    }
    if (NULL == (fp = fopen(outfile, "wb"))) {
        ret = 1;
        goto cleanup;
    }
    if (1 == *(char*)&platform_end) {
        platform_end = -1; /*little endian*/
    }
    fprintf(fp, "%s\n%d %d\n%d\n", (3==stride) ? "PF" : "Pf", (int) uwidth, (int) uheight, platform_end);
    byte_stride = stride * uwidth * sizeof(float);
    for (y = 0; y < uheight; y++ ) {
        if (fwrite((void *)((char*)img + byte_stride*y), sizeof(float)*stride, uwidth, fp) != uwidth) {
            ret = 1;
            goto cleanup;
        }
    }
cleanup:
    if (NULL != fp) {
        fclose(fp);
    }
    return ret;
}

